import os
import sys

from Cython.Build import cythonize

# use cythonize to build the extensions
modules = ["geomm/pyqcprot.pyx",]

extensions = cythonize(modules)

def build(setup_kwargs):
    """Needed for the poetry building interface."""

    setup_kwargs.update({
        'ext_modules' : extensions,
    })



import numpy as np

from geomm.centroid import centroid

def center(coords, idxs=None, weights=None):
    """Center coordinates at the origing based on their center of mass. If
    idxs are given the center of mass is computed only from those
    coordinates and if weights are given a weighted center of mass is
    computed.

    Inputs:

        coords :: the coordinates you wish to center

        idxs (optional) :: the idxs of the coordinates to actually
        compute the centroid on, although the translation will act on
        all the coordinates.

        weights (optional) :: give weights to the coordinates for a
           weighted centroid ('center of mass')

    """

    if idxs is None:
        centered_coords = coords - centroid(coords, weights=weights)
    else:
        centered_coords = coords - centroid(coords[idxs], weights=weights)

    return centered_coords




import numpy as np

def calc_rmsd(ref_coords, coords, idxs=None):

    # make sure the coords are the same size
    assert len(ref_coords.shape) == len(coords.shape), "rank of coords are not the same"
    assert ref_coords.shape[0] == coords.shape[0], \
        "Number of coordinates are not the same"

    # make sure the number of dimensions is 3
    assert (ref_coords.shape[1] == 3) and (coords.shape[1] == 3), \
        "Number of dimensions are not the same"

    if idxs is None:
        idxs = np.arange(ref_coords.shape[0])


    rmsd = np.sqrt(np.sum(np.square(coords[idxs, :] - ref_coords[idxs, :]),
                          axis=(0,1))/idxs.shape[0])

    return rmsd




import numpy as np

def centroid(coords, weights=None):
    """Return the centroid (AKA center of geometry or when weights
    correspond to masses the 'center of mass (COM)') of a set of
    coordinates.

    Inputs:

        coords :: the coordinates to find the centroid of

        weights (optional) :: give weights to the coordinates for a
           weighted centroid ('center of mass')

    """

    # no weights
    if weights is None:
        centroid = np.average(coords, axis=0)
    # use the weights
    else:
        centroid = np.average(coords, axis=0, weights=weights)

    return centroid



import numpy as np

from geomm.theobald_qcp import theobald_qcp
from geomm.centering import center
from geomm.centroid import centroid

def superimpose_traj(ref_coords, traj,
                centered=False, idxs=None, weights=None,
                rot_mat=False, rmsd=False):
    """Superimpose a whole trajectory onto a set of ref_coords
    (see superimpose for more info)"""
    
    return [superimpose(ref_coords, coords, centered=centered, idxs=idxs, weights=weights, rot_mat=rot_mat, rmsd=rmsd) for coords in traj]

def superimpose(ref_coords, coords,
                centered=False, idxs=None, weights=None,
                rot_mat=False, rmsd=False):
    """Superimpose a set of coordinates to reference coordinates using the
    Theobald-QCP method.

    Inputs:

      ref_coords :: the template coordinates that `coords` will be aligned to.

      coords :: the coordinates which will be aligned to the template
      coordinates and have the rotation matrix applied to them and
      returned.

      centered (optional) :: Default `False`, if False will center
      the coordinates to be superimposed to the origin to apply the
      rotation matrix.

      idxs (optional) :: Default `None`. If given will superimpose the
      coordinates based only on the alignment on this subset of atoms
      to the reference.

      rot_mat (optional) :: Default `False`. If True will return the
      rotation matrix computed from the call to Theobald-QCP.

      rmsd (optional) :: Default `False`. If True will return the rmsd
      computed from the call to Theobald-QCP. NOTE: this may not be
      exactly the same as the computation of the RMSD by the
      definition (i.e. the explicit equation implemented here in
      `geomm.rmsd.calc_rmsd`)

      weights (optional) :: give weights to the coordinates for a
      weighted centroid ('center of mass') used in translations to the
      origin and to the centroid of the reference structure.

    """

    # find the centroid of the reference coordinates
    if idxs is not None:
        ref_centroid = centroid(ref_coords[idxs], weights=weights)
    else:
        ref_centroid = centroid(ref_coords, weights=weights)

    # if the coordinates are not already centered we must center them
    if not centered:
        centered_coords = center(coords, idxs=idxs, weights=weights)
        centered_ref_coords = center(ref_coords, idxs=idxs, weights=weights)
    else:
        centered_coords = coords
        centered_ref_coords = ref_coords
    
    # perform the theobald_qcp method to get the rotation matrix
    qcp_rmsd, rotation_matrix = theobald_qcp(centered_ref_coords, centered_coords,
                                             idxs=idxs,
                                             rot_mat=True, weights=weights)

    # rotate coords according to the rotation matrix
    rot_coords = np.dot(centered_coords, rotation_matrix)

    # translate the rotated coordinates to the reference centroid
    sup_coords = rot_coords + ref_centroid

    # return results based on what was asked for
    if rot_mat and rmsd:
        return sup_coords, rotation_matrix, qcp_rmsd
    elif rot_mat:
        return sup_coords, rotation_matrix
    elif rmsd:
        return sup_coords, qcp_rmsd
    else:
        return sup_coords



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['geomm']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.28.5,<0.29.0', 'numpy>=1.15,<2.0']

setup_kwargs = {
    'name': 'geomm',
    'version': '0.1.1',
    'description': 'Simple and stupid library for strictly geometric (geom) operations (zero-knowledge of topology) common to computational biology and chemistry macromolecules (mm).',
    'long_description': '* GEOMM\n\nA simple no-nonsense library for computing common geometry on\nmacromolecular systems.\n\nThis library aims to work completely on numpy arrays and knows nothing\nabout molecule or system topologies, atom types, or other "real world"\nproperties.\n\nYou only need to provide atom coordinates, atom indices (for molecule,\ngroup, molecule selections etc.), and other parameters for functions.\n\nThus it aims to be more like numpy or scipy itself than a typical\nmacromolecular software library, except that the particular routines\nand functions are usually found in that context.\n\n** Installation\n\nTo install geomm make sure you have cython installed then:\n#+BEGIN_SRC bash\ngit clone https://github.com/ADicksonLab/geomm.git\ncd geomm\n\n# compile the cython modules\npython setup.py build_ext --inplace\n# install it\npip install -e .\n\n#+END_SRC\n\n** Community\n\nDiscussion takes place on riot.im (#geomm:matrix.org) which is a slack-like app that works\non the Matrix protocol:\n[[https://riot.im/app/#/room/#geomm:matrix.org]]\n\n\nYou can also contact me directly:\n\nsamuel.lotz@salotz.info\n',
    'author': 'Samuel Lotz',
    'author_email': 'samuel.lotz@salotz.info',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.7',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import numpy as np

def box_vectors_to_lengths_angles(box_vectors):
    """Convert box vectors for a single 'frame' to lengths and angles."""

    # calculate the lengths of the vectors through taking the norm of
    # them
    unitcell_lengths = []
    for basis in box_vectors:
        unitcell_lengths.append(np.linalg.norm(basis))
    unitcell_lengths = np.array(unitcell_lengths)

    # calculate the angles for the vectors
    unitcell_angles = np.array([np.degrees(
                        np.arccos(np.dot(box_vectors[i], box_vectors[j])/
                                  (np.linalg.norm(box_vectors[i]) * np.linalg.norm(box_vectors[j]))))
                       for i, j in [(0,1), (1,2), (2,0)]])

    return unitcell_lengths, unitcell_angles



import numpy as np

def group_pair_traj(traj, side_length_list, member_a_idxs, member_b_idxs):
    """Calls group_pair for each frame in traj"""
    return [group_pair(coords, unitcell_side_lengths, member_a_idxs, member_b_idxs)
            for coords,unitcell_side_lengths in zip(traj,side_length_list)]

def group_pair(coords, unitcell_side_lengths, member_a_idxs, member_b_idxs):
    """For a pair of group of coordinates (e.g. atoms) this moves member_b
    coordinates to the image of the periodic unitcell that minimizes
    the difference between the centers of geometry between the two
    members (e.g. a protein and ligand).

    """

    # take the difference between the average coords of each
    # molecule.
    unitcell_half_lengths = unitcell_side_lengths * 0.5

    # initialize a new array for the coordinates
    grouped_coords = np.copy(coords)

    # get the coordinates for the ligand and member_b separately
    member_a_coords = coords[member_a_idxs, :]
    member_b_coords = coords[member_b_idxs, :]

    # calculate the centroids for each member
    member_a_centroid = member_a_coords.mean(axis=0)
    member_b_centroid = member_b_coords.mean(axis=0)

    # calculate the difference between them from the given coordinates
    centroid_dist = member_a_centroid - member_b_centroid

    # now we need to move one relative to the other by unitcell
    # lengths in order to find one that makes the difference between
    # the centroids smallest

    # When the centroid_dist are larger than the unitcell half length in
    # both the positive and negative direction

    # The positive direction

    # this gives us a tuple of two arrays. THe first array is the
    # indices of the frames that satisfied the boolean expression. The
    # second array is the index of the dimension that satisfied the
    # boolean expression, i.e. 0,1,2 for x,y,z. For example if we have
    # a frame where member_b is outside the box on two dimensions we
    # could have (array([0,0]), array([0,2])) for frame 1 on both x
    # and z dimensions. and only (array([0]), array([0])) for outside
    # on the first frame x dimension.
    pos_idxs = np.where(centroid_dist > unitcell_half_lengths)[0]

    # we simply pair elements from each of those arrays (pairwise) to
    # iterate over them. Essentially reshaping the two arrays
    for dim_idx in pos_idxs:
        # change the ligand coordinates to recenter them by adding the
        # cube side length in that direction
        grouped_coords[member_b_idxs, dim_idx] = (member_b_coords[:, dim_idx] +
                                                unitcell_side_lengths[dim_idx])

    # the negative direction is the same as the positive direction
    # except for the boolean expression in the where command and we
    # subtract the cube side length instead of adding it
    neg_idxs = np.where(centroid_dist < -unitcell_half_lengths)[0]
    for dim_idx in neg_idxs:
        grouped_coords[member_b_idxs, dim_idx] = (member_b_coords[:, dim_idx] -
                                                  unitcell_side_lengths[dim_idx])


    return grouped_coords


def apply_rectangular_pbcs(coords, unitcell_side_lengths, center_point=(0., 0., 0.,)):
    """Apply rectangular Periodic Boundary Conditions (PBCs) given the
    lengths of the unitcell and a center point positions of the box in
    the coordinate space. The default for the center point is (0,0,0)
    which is the case for OpenMM MD frames but not other MD systems.

    """

    # check to make sure everything looks okay
    assert len(coords.shape) == 2, \
        "coordinates should be rank 2 array, "\
        "this function operates on individual frames not trajectories."
    assert coords.shape[1] == 3, "coordinates are not of 3 dimensions"
    assert len(center_point) == 3, "center point is not of 3 dimensions"
    assert len(unitcell_side_lengths) == 3, "Unitcell side lengths are not of dimension 3"

    # cast the center point to an array
    center_point = np.array(center_point)

    # Calculate half box sizes
    unitcell_half_lengths = unitcell_side_lengths * 0.5

    # initialize the coordinates to be wrapped
    wrapped_coords = np.copy(coords)

    # find coords which are outside the box in the positive direction
    pos_idxs = np.where(coords > center_point + unitcell_half_lengths)

    # Groups the frame_idx, atom_idx, and dim_idx
    pos_idxs = list(zip(pos_idxs[0], pos_idxs[1]))

    # Restrict particle coordinates to the simulation box
    for atom_idx, dim_idx in pos_idxs:
        wrapped_coords[atom_idx, dim_idx] = (coords[atom_idx, dim_idx] -
                                                 unitcell_side_lengths[dim_idx])

    # Find where coords are less than  the negative half box sizes
    neg_idxs = np.where(coords < center_point - unitcell_half_lengths)

    # Groups the fram_idx, atom_idx and dim_idx where they are greater
    # than half box sizes
    neg_idxs = list(zip(neg_idxs[0], neg_idxs[1]))

    # Restrict particle coordinates to the simulation box
    for atom_idx, dim_idx in neg_idxs:
        wrapped_coords[atom_idx, dim_idx] = (coords[atom_idx, dim_idx] +
                                                 unitcell_side_lengths[dim_idx])

    return wrapped_coords

def recenter_pair(coords, unitcell_side_lengths, member_a_idxs, member_b_idxs):
    """
    This method moves group of ligand and member_b to the given new
    center. The geometric mean is used to find the center point of
    ligand and receptor group.

    """

    # group the pair
    grouped_coords = group_pair(coords, unitcell_side_lengths,
                                member_a_idxs, member_b_idxs)

    # then recenter them as a complex
    complex_idxs = (member_a_idxs, member_b_idxs)
    recentered_coords = recenter_complex(grouped_coords, unitcell_side_lengths, complex_idxs)

    return recentered_coords

def recenter_complex(coords, unitcell_side_lengths, complex_idxs):
    """Recenter a periodic unitcell around a complex (a list of lists of
    atoms idxs for each member of the complex), by first computing the
    centroids of each member then computing the centroid of the
    centroids, and using that as the new center of the box.

    """

    # compute the centroids of each member in the complex
    member_centroids = []
    for member_idxs in complex_idxs:
        centroid = coords[member_idxs].mean(axis=0)
        member_centroids.append(centroid)
    member_centroids = np.array(member_centroids)

    # compute the centroid of the centroids
    complex_centroid = member_centroids.mean(axis=0)

    # apply the periodic boundary conditions around the complex
    # centroid to recenter
    recentered_coords = apply_rectangular_pbcs(coords, unitcell_side_lengths,
                                               center_point=complex_centroid)

    return recentered_coords


def group_complex(coords, unitcell_side_lengths, complexes_idxs):
    """Given several groups of atom idxs (say molecules) put them in the
    image which minimizes the distance between the individual
    centroids to the collective centroid.

    """

    raise NotImplementedError

    unitcell_half_lengths = unitcell_side_lengths * 0.5

    grouped_coords = np.copy(coords)

    # get the coordinates of each complex
    complex_coords = []
    for complex_idxs in complexes_idxs:
        complex_coords.append(coords[complex_idxs, :])

    return grouped_coords





import numpy as np

from geomm.pyqcprot import CalcRMSDRotationalMatrix

def theobald_qcp(ref_coords, coords, idxs=None, rot_mat=True, weights=None):
    """Wrapper around the pyqcprot implementation of the Theobald-QCP
    method for the calculation of RMSD and the RMSD minimizing rotation
    matrix. This function just gives a more pythonic API to the
    function

    Arguments:

    ref_coords :: the refence coordinates that will be aligned to

    coords :: the coordinates that will be rotated to match ref_coords

    rot_mat :: if True will return both the rmsd and the rotation matrix

    weights :: if your coordinates are weighted (e.g. mass) this is an array of those weights

    Returns:

    if rot_mat is True:

       rmsd :: (float) the rmsd of the two sets of coordinates

    if rot_mat is False:

       rmsd, rotation_matrix :: the rotation matrix that minimizes the RMSD

    """

    # make sure the coords are the same size
    assert ref_coords.shape[0] == coords.shape[0], \
        "Number of coordinates are not the same"

    # make sure the number of dimensions is 3
    assert (ref_coords.shape[1] == 3) and (coords.shape[1] == 3), \
        "Number of dimensions are not the same"

    # copy the coordinates to make sure we are not modifying them in
    # place
    ref_coords_copy = ref_coords.copy()
    coords_copy = coords.copy()

    # if idxs were given we use just those for aligning
    if idxs is not None:
        align_ref_coords = ref_coords_copy[idxs]
        align_coords = coords_copy[idxs]
    else:
        align_ref_coords = ref_coords_copy
        align_coords = coords_copy

    # the number of coordinates (atoms)
    n_coords = align_ref_coords.shape[0]

    # make sure the weights if given are the right size
    if weights is not None:
        assert weights.shape[0] == n_coords, \
            "Number of weights given does not match the number of coordinates"

    # make an empty rotation matrix to fill from the cython function,
    # this is always an array of length 9
    rotation_matrix = np.zeros((9,), dtype=np.float64)

    # calculate the rmsd, and the rotation matrix which is modified in-place
    rmsd = CalcRMSDRotationalMatrix(align_ref_coords, align_coords,
                                    n_coords, rotation_matrix, weights)

    # if the rotation matrix was asked for return it, otherwise don't
    if rot_mat:
        # reshape the rotation matrix to be 2D
        return rmsd, rotation_matrix.reshape( (3, 3) )
    else:
        return rmsd
